from typing import Any, Callable, Dict, Optional, Tuple

from PIL.Image import Image as PilImage

__all__ = ["DictionaryTransforms", "DictionaryTransform"]


class _DictionaryTransformBase(Callable):
    """
    _DictionaryTransformBase.

    A wrapper of Albumentations-style keyword args transforms. Compatible with
    torchvision transforms.
    """

    def __init__(
        self,
        preprocessing: Callable[[PilImage, Any], Dict[str, Any]],
        dictionary_transform: Callable[..., Dict[str, Any]],
        postprocessing: Callable[
            [Dict[str, Any]], Tuple[PilImage, Optional[Dict[str, Any]]]
        ],
    ):
        """
        __init__

        Args:
            preprocessing (Callable[[PilImage, Any], Dict[str, Any]]): Convert any target type into a dict
            dictionary_transform (Callable[..., Dict[str, Any]]): Only kwargs as inputs and then outputs a dict
            postprocessing (Callable[ [Dict[str, Any]], Tuple[PilImage, Optional[Dict[str, Any]]] ]):
                Converts dict outputs to torchvision (image, target) style
        """

        self.preprocessing = preprocessing
        self.core = dictionary_transform
        self.postprocessing = postprocessing

    def __call__(self, *args):
        out = self.preprocessing(*args)
        out = self.core(**out)  # always outputs an dict
        out = self.postprocessing(out)
        return out

    def __repr__(self):
        return repr(self.core)

    __str__ = __repr__


class DictionaryTransform(_DictionaryTransformBase):
    def __call__(self, image: PilImage) -> Any:
        return super().__call__(image)


class DictionaryTransforms(_DictionaryTransformBase):
    def __call__(self, image: PilImage, target: Any) -> Tuple[Any, Any]:
        return super().__call__(image, target)



import os
import os.path
import pathlib
from typing import Any, Callable, Dict, List, Optional, Tuple

import PIL
from PIL.Image import Image as PilImage
from torchvision.datasets.utils import download_and_extract_archive, verify_str_arg
from torchvision.datasets.vision import VisionDataset

__all__ = ["StanfordDogs"]


class StanfordDogs(VisionDataset):
    """
    `Stanford Dogs Dataset.

    <http://vision.stanford.edu/aditya86/ImageNetDogs/>`_.

    Implementation of this dataset class was inspired by <https://github.com/zrsmithson/Stanford-dogs>.

    Args:
        root (str): Root directory of the dataset.
        split (str, optional): The dataset split, supports ``"train"`` (default) or ``"test"``.
        transforms (callable, optional): A function/transform that takes in a dictionary of
            `image` (PIL Image), `bboxes` (e.g. [[x0, y0, x1, y1]]), `labels` ([1]) and returns a transformed version.
        download (bool): Description of parameter `download`. Defaults to False.

    Attributes:
        classes (List[str]): Description of parameter `classes`.
        class_to_idx (Dict[str, int]): Description of parameter `class_to_idx`.
        OUTPUT_FIELDS (Tuple[str]): Description of parameter `OUTPUT_FIELDS: Tuple[str]`.
    """

    OUTPUT_FIELDS: Tuple[str] = (
        "image",
        "labels",
        "bboxes",
    )

    _RESOURCES = (
        (
            "http://vision.stanford.edu/aditya86/ImageNetDogs/images.tar",
            "1bb1f2a596ae7057f99d7d75860002ef",
        ),
        (
            "http://vision.stanford.edu/aditya86/ImageNetDogs/annotation.tar",
            "4298cc0031f6bc6e74612ac83b5988e2",
        ),
        (
            "http://vision.stanford.edu/aditya86/ImageNetDogs/lists.tar",
            "edbb9f16854ec66506b5f09b583e0656",
        ),
    )

    def __init__(
        self,
        root: str,
        split: str = "train",
        transforms: Optional[Callable] = None,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ):

        self._split = verify_str_arg(split, "split", ("train", "test"))

        super().__init__(
            root,
            transforms=transforms,
            transform=transform,
            target_transform=target_transform,
        )

        self._base_folder = pathlib.Path(self.root) / "stanford-dogs"
        self._image_folder = self._base_folder / "Images"
        self._anno_folder = self._base_folder / "Annotation"
        self._list_folder = self._base_folder

        if download:
            self._download()

        if not self._check_exists():
            raise RuntimeError(
                "Dataset not found. You can use download=True to download it"
            )

        self._images = []  # list of relative paths to images
        self._labels = []  # list of int labels (zero-based)
        self._annos = []  # list of relative paths to annotations
        self._load_split()  # populates _images, _labels, _annos

        self.classes = [
            " ".join(part.title() for part in raw_cls.split("_"))
            for raw_cls, _ in sorted(  # sort the set of (str, int) using int
                {
                    (
                        _image.split("/")[0].split("-", 1)[1],
                        label,
                    )  # split '-' from left up to 1 '-'
                    for _image, label in zip(self._images, self._labels)
                },
                key=lambda image_id_and_label: image_id_and_label[1],
            )
        ]
        # e.g. n02085620-Chihuahua/n02085620_2650.jpg,
        #      n02095314-wire-haired_fox_terrier/n02095314_3052.jpg

        self.class_to_idx = dict(zip(self.classes, range(len(self.classes))))

        self._images = [self._image_folder / i for i in self._images]
        self._annos = [self._anno_folder / a for a in self._annos]

        if len(self._images) != len(self._annos):
            raise ValueError(
                f"Number of images ({len(self._images)}) is not consistent with number of annotations ({len(self._annos)})"
            )

        # list of List[List[int]], e.g. [[x0, y0, x1, y1], [x0, y0, x1, y1], ...]
        self._boxes = []
        self._load_boxes()

    def __len__(self) -> int:
        return len(self._images)

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        image: PilImage = PIL.Image.open(self._images[idx]).convert("RGB")
        boxes: List[List[int]] = self._boxes[idx]
        label: int = self._labels[idx]
        # the number of labels per image is the same as number of bboxes
        labels: List[int] = [label] * len(boxes)  #

        target = {"bboxes": boxes, "labels": labels}

        if self.transforms:
            image, target = self.transforms(image, target)

        return image, target

    def _download(self) -> None:
        if self._check_exists():
            return

        for url, md5 in self._RESOURCES:
            download_and_extract_archive(
                url, download_root=str(self._base_folder), md5=md5
            )

    def _check_exists(self) -> bool:
        for folder in (self._image_folder, self._anno_folder, self._list_folder):
            if not (os.path.exists(folder) and os.path.isdir(folder)):
                return False
        else:
            return True

    def _load_split(self):
        import scipy.io

        split_info = scipy.io.loadmat(self._list_folder / f"{self._split}_list.mat")
        self._images = [f[0][0] for f in split_info["file_list"]]
        self._labels = [l[0] - 1 for l in split_info["labels"]]
        self._annos = [a[0][0] for a in split_info["annotation_list"]]

    def __load_boxes_per_image(self, path: str) -> List[int]:
        import xml.etree.ElementTree

        e = xml.etree.ElementTree.parse(path).getroot()
        boxes = []
        for objs in e.iter("object"):
            boxes.append(
                [
                    int(objs.find("bndbox").find("xmin").text),
                    int(objs.find("bndbox").find("ymin").text),
                    int(objs.find("bndbox").find("xmax").text),
                    int(objs.find("bndbox").find("ymax").text),
                ]
            )
        return boxes

    def _load_boxes(self):
        for anno in self._annos:
            self._boxes.append(self.__load_boxes_per_image(anno))



from torchvision.datasets import *
from torchvision.datasets.vision import StandardTransform

from .oxford_iiit_pet_plus import OxfordIIITPetPlus
from .stack import StackDataset
from .stanford_dogs import StanfordDogs



import copy
from typing import *

import torch
import torch.nn as nn
from torch import Tensor
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor

__all__ = ["MultiVocabularyFastRCNNPredictor"]


class MultiVocabularyFastRCNNPredictor(FastRCNNPredictor):
    """
    MultiVocabularyFastRCNNPredictor.

    Train a FasterRCNN using multiple Vocabularies Set vocabulary using
    `set_vocabulary` Export partial head for a single vocabulary using
    `export_current_vocab_heads`
    """

    def __init__(self, in_channels, vocabulary_sizes: Dict[str, int]):
        vocabulary_sizes = {str(k): int(v) for k, v in vocabulary_sizes.items()}
        torch._assert(
            all([v > 0 for v in vocabulary_sizes.values()]),
            f"Vocabulary size should be more than one but got {vocabulary_sizes}",
        )
        total_num_classes = 1 + sum((v for v in vocabulary_sizes.values()))
        # __background__ is a default class with index 0
        super().__init__(in_channels, total_num_classes)
        self.total_num_classes = total_num_classes
        self._vocabulary_sizes = vocabulary_sizes
        vocab_to_range = {}  # '__background__' is 0
        current_index = 0
        for vocab, size in vocabulary_sizes.items():
            vocab_to_range[vocab] = (current_index + 1, current_index + size)
            current_index += size
        self._vocabulary_to_range = vocab_to_range
        self.__current_vocabulary = "all"

    def extra_repr(self) -> str:
        return f"current_vocabulary: {self.current_vocabulary}"

    def _is_in_vocab(self, label: int, vocab_name: str) -> bool:
        min_idx, max_idx = self._vocabulary_to_range[vocab_name]
        return min_idx <= label <= max_idx

    def set_vocabulary(self, vocab_name: str = "all") -> None:
        if vocab_name != "all":
            torch._assert(
                vocab_name in self._vocabulary_sizes,
                f"Vocabulary `{vocab_name}` not found.",
            )
        self.__current_vocabulary = vocab_name

    @property
    def label_shift(self) -> int:
        return self._vocabulary_to_range[self.current_vocabulary][0]

    @property
    def current_vocabulary(self) -> str:
        return self.__current_vocabulary

    def forward(self, x) -> Tuple[Tensor, Tensor]:
        logits, box_deltas = super().forward(x)
        if self.current_vocabulary == "all":
            return logits, box_deltas

        min_idx, max_idx = self._vocabulary_to_range[self.current_vocabulary]
        logits[:, 1:min_idx] += float("-inf")
        logits[:, max_idx + 1 :] += float("-inf")

        return logits, box_deltas

    def export_current_vocab_heads(self) -> FastRCNNPredictor:

        in_channels = self.cls_score.in_features
        cls_w = copy.deepcopy(self.cls_score.weight).cpu()
        cls_b = copy.deepcopy(self.cls_score.bias).cpu()
        box_w = copy.deepcopy(self.bbox_pred.weight).cpu()
        box_b = copy.deepcopy(self.bbox_pred.bias).cpu()

        if self.current_vocabulary == "all":
            m = FastRCNNPredictor(in_channels, self.total_num_classes)
        else:
            min_idx, max_idx = self._vocabulary_to_range[self.current_vocabulary]
            m = FastRCNNPredictor(
                in_channels, 1 + self._vocabulary_sizes[self.current_vocabulary]
            )

            cls_w = cls_w[[0] + [i for i in range(min_idx, max_idx + 1)]]
            cls_b = cls_b[[0] + [i for i in range(min_idx, max_idx + 1)]]
            box_w = box_w[
                [i for i in range(4)] + [i * 4 for i in range(min_idx, max_idx + 1)]
            ]
            box_b = box_b[
                [i for i in range(4)] + [i * 4 for i in range(min_idx, max_idx + 1)]
            ]

        m.cls_score.weight = nn.Parameter(cls_w)
        m.cls_score.bias = nn.Parameter(cls_b)
        m.bbox_pred.weight = nn.Parameter(box_w)
        m.bbox_pred.bias = nn.Parameter(box_b)

        return m



import random
from typing import *

from torch.utils.data import Dataset

__all__ = ["StackDataset"]


class StackDataset(Dataset):

    _repr_indent = 4

    def __init__(self, datasets: Dict[str, Dataset], pad_short: bool = True):
        super().__init__()
        self.pad_short = pad_short
        assert isinstance(
            datasets, dict
        ), f"Input `datasets` should be a dict but got type {type(datasets)}"
        assert (
            len(datasets) > 1
        ), f"Input `datasets` should be at least 2 but got {len(datasets)}"

        self.datasets = datasets
        max_len = 0
        min_len = float("inf")
        data_len = {}
        for name, ds in datasets.items():
            max_len = max(len(ds), max_len)
            min_len = min(len(ds), min_len)
            data_len[name] = len(ds)
        self.min_len = max(0, min_len)
        self.max_len = max(self.min_len, max_len)
        self.data_len = data_len
        self._num_output_args = {
            vocab: len(self.datasets[vocab][0]) for vocab in self.datasets
        }

    def __getitem__(self, idx):

        if idx < 0:
            jdx = len(self) + idx
            if jdx < 0:
                raise IndexError(
                    f"Index {idx} is out of bound [{-len(self)}, {len(self)})."
                )
            idx = jdx
        if idx >= len(self):
            raise IndexError(f"Index {idx} >= {len(self)} is out of bound.")

        outputs = {}

        for name, ds in self.datasets.items():
            if idx >= self.data_len[name]:
                jdx = random.randint(0, self.data_len[name] - 1)
            else:
                jdx = idx
            outputs[name] = ds[jdx]

        return outputs

    def __len__(self) -> int:
        if self.pad_short:
            return self.max_len
        else:
            return self.min_len

    def __repr__(self) -> str:
        out = [
            "StackDataset",
            " " * self._repr_indent + f"`pad_short`: {self.pad_short}",
            " " * self._repr_indent + f"`len`: {len(self)}",
        ]
        for name, ds in self.datasets.items():
            out.append(f"`{name}`:")
            r = repr(ds)
            for l in r.split("\n"):
                out.append(" " * self._repr_indent + f"{l}")

        return "\n".join(out)

    def collate_fn(
        self, batch: List[Dict[str, Tuple[Any]]]
    ) -> Dict[str, Tuple[List[Any]]]:
        """
        `collate_fn` for `StackDataset` items.

        Args:
            batch (List[Dict[str, Tuple[Any]]]): A list of dict items, i.e. vocab name -> data item

        Returns:
            Dict[str, Tuple[List[Any]]]: A dict of batched items, i.e. vocab name -> tuple of args where args are List[Any]
        """

        batched = {
            vocab: [[] for _ in range(n)] for vocab, n in self._num_output_args.items()
        }

        for item in batch:
            for vocab, n in self._num_output_args.items():
                for adx in range(n):
                    batched[vocab][adx].append(item[vocab][adx])

        return batched



from torchvision.transforms import Compose as _Compose

__all__ = ["Compose"]


class Compose(_Compose):
    """
    Compose.

    A better version of torchvision Compose.
    """

    def __call__(self, *args):
        for t in self.transforms:
            args = t(*args)

        args = tuple(args)

        return args if len(args) > 1 else args[0]



from typing import Any, Callable, Dict, Optional, Tuple

import numpy as np
import torch
import torchvision.transforms.functional as TF
from PIL.Image import Image as PilImage

__all__ = ["ToTensors"]


class ToTensors(Callable):
    """
    ToTensors.

    A better version of torchvision `ToTensor` which takes only one PIL image
    argument. `ToTensors` can convert any input types to torch.Tensor
    """

    def __init__(self, vocabulary: Optional[Dict[str, int]] = None):

        self.vocabulary = vocabulary

    def __call__(self, *args) -> Tuple[Any]:
        out = tuple([self._to_tensor(a) for a in args])
        return out if len(out) > 1 else out[0]

    def _to_tensor(self, x: Any):

        if isinstance(x, np.ndarray):
            return torch.tensor(x)

        if isinstance(x, (list, tuple, set)):

            as_list = [self._to_tensor(y) for y in x]
            # the elements are the same type `int` or `float`
            if all([isinstance(y, int) for y in as_list]):
                return torch.tensor(as_list).long()
            if all([isinstance(y, float) for y in as_list]):
                return torch.tensor(as_list).float()

            if all([isinstance(y, (list, tuple)) for y in x]):
                _dtype_size = None
                for item in as_list:
                    if not isinstance(item, torch.Tensor):
                        break
                    if _dtype_size:
                        _dt, _sz = _dtype_size
                        if _dt == item.dtype and _sz == item.size():
                            continue
                        else:
                            break
                    else:
                        _dtype_size = item.dtype, item.size()
                else:
                    return torch.stack(as_list, dim=0)

            # otherwise return tensors in the original container type
            return type(x)(as_list)

        if isinstance(x, dict):
            return {k: self._to_tensor(v) for k, v in x.items()}

        if isinstance(x, str):
            return self.vocabulary[x]

        if isinstance(x, (int, float)):
            return x

        if isinstance(x, PilImage):
            return TF.to_tensor(x)

        return x

    def __repr__(self) -> str:
        head = self.__class__.__name__
        body = f"(vocabulary={self.vocabulary})"
        return head + body



from typing import Sequence, Union

__all__ = ["Vocabulary"]


class Vocabulary:
    """
    Vocabulary.

    Use `__getitem__` operator to fetch class name or class index
    """

    def __init__(self, name: str, class_names: Sequence[str]):
        """
        __init__

        Args:
            name (str): Vocabulary name
            class_names (Sequence[str]): A sequence of class names
        """
        self.name = str(name)
        self.class_names = tuple([str(n) for n in class_names])
        self.name_to_idx = {c: i for i, c in enumerate(self.class_names)}

    def __getitem__(self, key_or_idx: Union[int, str]):
        if isinstance(key_or_idx, str):
            return self.name_to_idx[key_or_idx]
        elif isinstance(key_or_idx, int):
            return self.class_names[key_or_idx]
        else:
            raise TypeError(f"Key can be `int` or `str` but got `{type(key_or_idx)}`")

    def __len__(self):
        return len(self.class_names)

    def __repr__(self):
        head = self.__class__.__name__
        body = []
        body.append("(")
        body.append(f"name={repr(self.name)}")
        body.append(", ")
        body.append(f"class_names={repr(self.class_names)}")
        body.append(")")
        return head + "".join(body)

    __str__ = __repr__



from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Union

import numpy as np
import PIL
from torchvision.datasets import OxfordIIITPet

__all__ = ["OxfordIIITPetPlus"]


class OxfordIIITPetPlus(OxfordIIITPet):
    """`Oxford-IIIT Pet Dataset   <https://www.robots.ox.ac.uk/~vgg/data/pets/>`_.
    Args:
        root (string): Root directory of the dataset.
        split (string, optional): The dataset split, supports ``"trainval"`` (default) or ``"test"``.
        target_types (string, sequence of strings, optional): Types of target to use. Can be ``category`` (default) or
            ``segmentation``. Can also be a list to output a tuple with all specified target types. The types represent:
                - ``category`` (int): Label for one of the 37 pet categories.
                - ``coarse_category`` (int): Label for cat or dog
                - ``segmentation`` (PIL image): Segmentation trimap of the image.
                - ``detection``: Bounding box around segmentation mask
            If empty, ``None`` will be returned as target.
        transform (callable, optional): A function/transform that takes in a PIL image and returns a transformed
            version. E.g, ``transforms.RandomCrop``.
        target_transform (callable, optional): A function/transform that takes in the target and transforms it.
        download (bool, optional): If True, downloads the dataset from the internet and puts it into
            ``root/oxford-iiit-pet``. If dataset is already downloaded, it is not downloaded again.
    """

    def __init__(
        self,
        root: str,
        split: str = "trainval",
        target_types: Union[Sequence[str], str] = "category",
        transforms: Optional[Callable] = None,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download: bool = False,
    ):

        super().__init__(
            root,
            split,
            ("category", "segmentation"),
            transforms,
            transform,
            target_transform,
            download,
        )

        self.read_me = self._load_readme()
        # load species and breed labels
        self._load_labels()
        self._compute_boxes()
        self.target_types = target_types

    def _compute_boxes(self):
        self._boxes = []
        for idx in range(len(self)):
            mask = PIL.Image.open(self._segs[idx])
            ydx, xdx = np.where(np.array(mask) == 3)
            box = [[xdx.min(), ydx.min(), xdx.max(), ydx.max()]]
            self._boxes.append(box)

    def _load_readme(self) -> str:
        with open(self._anns_folder / "README", "r") as f:
            return f.read()

    def _load_labels(self) -> None:

        self.class_to_idx_tuple = {}
        self.idx_to_name = {}
        self.idx_to_species_breed_idx = {}

        with open(self._anns_folder / "list.txt", "r") as f:
            lines = f.readlines()

        for line in lines:
            if line.startswith("#"):
                continue
            fn, general, coarse, fine = line.split(" ")

            name = " ".join((w.capitalize() for w in fn.lower().split("_")[:-1]))
            if name not in self.class_to_idx_tuple:
                if int(coarse) - 1 == 0:
                    coarse_name = "Cat"
                elif int(coarse) - 1 == 1:
                    coarse_name = "Dog"
                else:
                    raise ValueError(
                        f"Coarse label should only be 1 or 2 but got {coarse}"
                    )

                name = f"{name}, {coarse_name}"

                self.class_to_idx_tuple[name] = (
                    int(general) - 1,
                    int(coarse) - 1,
                    int(fine) - 1,
                )
                self.idx_to_name[int(general) - 1] = name
                self.idx_to_species_breed_idx[int(general) - 1] = (
                    int(coarse) - 1,
                    int(fine) - 1,
                )

        self.species_to_species_idx = {"Cat": 0, "Dog": 1}
        self.cat_breed_to_breed_idx = {}
        self.dog_breed_to_breed_idx = {}

        for name, idx_tuple in self.class_to_idx_tuple.items():
            breed, species = name.split(", ")
            if species == "Cat":
                self.cat_breed_to_breed_idx[breed] = idx_tuple[-1]
            else:  # species == 'Dog'
                self.dog_breed_to_breed_idx[breed] = idx_tuple[-1]

    def __getitem__(self, idx: int) -> Tuple[Any, Dict[str, Any]]:
        image = PIL.Image.open(self._images[idx]).convert("RGB")

        target: Dict[str, Any] = {}
        for target_type in self._target_types:
            if target_type == "category":
                target["flat_label"] = self._labels[idx]
            elif target_type == "coarse_category":
                species_label, breed_label = self.idx_to_species_breed_idx[
                    self._labels[idx]
                ]
                target["flat_label"] = self._labels[idx]
                target["species_label"] = species_label
                target["breed_label"] = breed_label
            elif target_type == "detection":
                target["boxes"] = self._boxes[idx]
            elif target_type == "segmentation":
                target.append(PIL.Image.open(self._segs[idx]))
            else:
                raise KeyError(f"Target type ({target_type}) not found.")

        if self.transforms:
            image, target = self.transforms(image, target)

        return image, target
